/**
 * ModServ IRC Services
 * Version 2009
 * Copyright (C) 2007, 2008 Christian Jurk <commx@commx.ws>
 * All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 3 of the License,
 * or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <http://www.gnu.org/licenses/>.
 */

#include "control.h"
#include <MS_Main.h>
#include <MS_Channels.h>
#include <MS_Flagops.h>

void ModControl::c_get(MUser *sender, const QStringList &strlArgs) {
  if (strlArgs.size() < 2) {
    __ModControl->sendnotice(sender, QString("Usage: get var"));
    return;
  }

  QString value = globalMain->config->get(strlArgs.at(1), QString(""));
  if (value.isEmpty()) {
    __ModControl->sendnotice(sender, QString("No configuration parameter \"%1\" set.").arg(strlArgs.at(1)));
  } else {
    __ModControl->sendnotice(sender, QString("Configuration parameter \"%1\" = \"%2\"").arg(strlArgs.at(1)).arg(value));
  }
}

void ModControl::c_help(MUser *sender, const QStringList &strlArgs) {
  if (strlArgs.size() < 2) {
    __ModControl->sendnotice(sender, QString("Usage: help command"));
    return;
  }

  foreach(cmdHandler *hdl, __ModControl->getHandlers()) {
    if (hdl->command == strlArgs.at(1).toLower()) {
      if (!hdl->helpmsg.isEmpty()) {
        QStringList strlMsg;
        if (hdl->helpmsg.contains("\n"))
          strlMsg = hdl->helpmsg.split("\n");
        else
          strlMsg << hdl->helpmsg;

        __ModControl->sendnotice(sender, QString("%1 - %2").arg(hdl->command).arg(hdl->description));
        foreach(QString strMsg, strlMsg) {
          __ModControl->sendnotice(sender, strMsg);
        }
      } else {
        __ModControl->sendnotice(sender, QString("Sorry, there is no help available for command %1").arg(hdl->command));
      }

      return;
    }
  }

  __ModControl->sendnotice(sender, QString("Unknown command %1.").arg(strlArgs.at(1)));
}

void ModControl::c_insmod(MUser *sender, const QStringList &strlArgs) {
  if (strlArgs.size() < 2) {
    __ModControl->sendnotice(sender, QString("Usage: insmod path_to_module"));
    return;
  }

  MModule *mod;
  foreach(mod, modList) {
    if (mod->filePath() == strlArgs.at(1)) {
      __ModControl->sendnotice(sender, QString("Module at \"%1\" (%2) is already loaded.").arg(mod->filePath().absoluteFilePath()).arg(mod->name()));
      return;
    }
  }

  mod = new MModule(strlArgs.at(1));
  if (!globalDlError.isEmpty()) {
    __ModControl->sendnotice(sender, QString("Loading module \"%1\" failed: %2").arg(strlArgs.at(1)).arg(globalDlError));
    return;
  } else {
    __ModControl->sendnotice(sender, QString("Loading module \"%1\" OK, getting module parameters...").arg(strlArgs.at(1)));
    __ModControl->sendnotice(sender, QString("Name: %1, Version: %2, Build date: %3").arg(mod->name()).arg(mod->version()).arg(mod->built()));
  }
}

void ModControl::c_lsmod(MUser *sender, const QStringList &strlArgs) {
  if (strlArgs.size() < 0)
    return;

  int i = 1;
  foreach(MModule *mod, modList) {
    __ModControl->sendnotice(sender, QString("%1. Module \"%2\", Version %3 (Build date %4)").arg(i++).arg(mod->name()).arg(mod->version()).arg(mod->built()));
    __ModControl->sendnotice(sender, QString("   File system path: %1").arg(mod->filePath().absoluteFilePath()));
    __ModControl->sendnotice(sender, QString("   Loaded at %1").arg(mod->loadTime().toString("dd.MM.yyyy hh:mm")));
  }

  __ModControl->sendnotice(sender, QString("End of list."));
}

void ModControl::c_showcommands(MUser *sender, const QStringList &strlArgs) {
  if (strlArgs.size() == 0)
    return;

  /* prepare the sorted command list */
  QStringList strlCommands;
  QHash<QString, QString> hashDesc;

  foreach(cmdHandler *hdl, __ModControl->getHandlers()) {
    strlCommands << hdl->command;
    hashDesc.insert(hdl->command, hdl->description);
  }

  strlCommands.sort();
  for(int i=0; i < strlCommands.size(); i++) {
    __ModControl->sendnotice(sender, QString("%1 - %2").arg(strlCommands.at(i)).arg(hashDesc.value(strlCommands.at(i))));
  }

  __ModControl->sendnotice(sender, QString("End of list."));
}

void ModControl::c_rlmod(MUser *sender, const QStringList &strlArgs) {
  if (strlArgs.size() < 2) {
    __ModControl->sendnotice(sender, QString("Usage: rlmod path_to_module"));
    return;
  }

  MModule *mod;
  foreach(mod, modList) {
    if (mod->filePath() == strlArgs.at(1)) {
      if (mod->isSysMod()) {
        __ModControl->sendnotice(sender, QString("Module \"%1\" is a system module and cannot be un-/reloaded.").arg(mod->name()));
        return;
      }

      __ModControl->sendnotice(sender, QString("Attempting to unload module \"%1\"...").arg(mod->name()));
      delete mod;

      mod = new MModule(strlArgs.at(1));
      if (!globalDlError.isEmpty()) {
        __ModControl->sendnotice(sender, QString("Loading module \"%1\" failed: %2").arg(strlArgs.at(1)).arg(globalDlError));
        return;
      } else {
        __ModControl->sendnotice(sender, QString("Loading module \"%1\" OK, getting module parameters...").arg(strlArgs.at(1)));
        __ModControl->sendnotice(sender, QString("Name: %1, Version: %2, Build date: %3").arg(mod->name()).arg(mod->version()).arg(mod->built()));

      }

      return;
    }
  }
}

void ModControl::c_rmmod(MUser *sender, const QStringList &strlArgs) {
  if (strlArgs.size() < 2) {
    __ModControl->sendnotice(sender, QString("Usage: rmmod path_to_module"));
    return;
  }

  MModule *mod;
  foreach(mod, modList) {
    if (mod->filePath() == strlArgs.at(1)) {
      if (mod->isSysMod()) {
        __ModControl->sendnotice(sender, QString("Module \"%1\" is a system module and cannot be unloaded.").arg(mod->name()));
        return;
      }

      __ModControl->sendnotice(sender, QString("Attempting to unload module \"%1\"...").arg(mod->name()));
      delete mod;
      __ModControl->sendnotice(sender, QString("Done."));
      return;
    }
  }

  __ModControl->sendnotice(sender, QString("Unable to locate module \"%1\".").arg(strlArgs.at(1)));
}

void ModControl::c_serverlist(MUser *sender, const QStringList &strlArgs) {
  if (strlArgs.size() == 0)
    return;

  MServer *srv;
  int index = 1;
  QHash<unsigned int, MServer *>::iterator it;
  for(it = serverList.begin(); it != serverList.end(); ++it) {
    srv = it.value();

    QStringList strlSrvFlags;
    unsigned short sflags = srv->flags();
    if (FLAG_SET(sflags, MServer::FL_HIDDEN))
      strlSrvFlags << "Hidden";
    if (FLAG_SET(sflags, MServer::FL_SERVICES))
      strlSrvFlags << "Services";
    if (uplinkServer == srv)
      strlSrvFlags << "Uplink";
    if (strlSrvFlags.size() == 0)
      strlSrvFlags << "None";

    QString srvStatus;
    if (srv->state() == MServer::STATE_CONNECTED)
      srvStatus = "Connected, awaiting End of Burst";
    else if (srv->state() == MServer::STATE_END_OF_BURST)
      srvStatus = "Connected, awaiting End of Burst acklowledge";
    else if (srv->state() == MServer::STATE_LINKED)
      srvStatus = "Link syncronized";
    else
      srvStatus = "Unknown";

    __ModControl->sendnotice(sender, QString("%1. \2%2\2 numeric %3 (%4)").arg(index++).arg(srv->name()).arg(srv->numeric64()).arg(srv->description()));
    __ModControl->sendnotice(sender, QString("   Flags: %1, State: %2, Boot/Link timestamp: %3 / %4")
                                       .arg(strlSrvFlags.join(", "))
                                       .arg(srvStatus)
                                       .arg(srv->bootTS().toString("dd/MM/yy hh:mm:ss"))
                                       .arg(srv->linkTS().toString("dd/MM/yy hh:mm:ss")));
    __ModControl->sendnotice(sender, QString("   Usercount: %1").arg(srv->userList.size()));
  }
}

void ModControl::c_set(MUser *sender, const QStringList &strlArgs) {
  if (strlArgs.size() < 3) {
    __ModControl->sendnotice(sender, QString("Usage: set var value"));
    return;
  }

  QStringList strlVal;
  for(int i=2; i < strlArgs.size(); i++)
    strlVal << strlArgs.at(i);

  globalMain->config->set(strlArgs.at(1), strlVal.join(""));
  __ModControl->sendnotice(sender, QString("Done."));
}

void ModControl::c_showchan(MUser *sender, const QStringList &strlArgs) {
  if (strlArgs.size() < 2) {
    __ModControl->sendnotice(sender, QString("Usage: showchan #channel"));
    return;
  }

  MChannel *chan;
  if (!(chan = MChannel::find(strlArgs.at(1)))) {
    __ModControl->sendnotice(sender, QString("Unknown channel %1.").arg(strlArgs.at(1)));
    return;
  }

  QString strmode;
  QHash<char, quint32>::iterator it;
  for(it = cmodeList.begin(); it != cmodeList.end(); ++it) {
    if (FLAG_SET(chan->modeflags(), it.value()))
      strmode.append(it.key());
  }

  if (strmode.contains("k") && !chan->key().isEmpty())
    strmode.append(QString(" %1").arg(chan->key()));
  if (strmode.contains("l") && chan->limit() > 0)
    strmode.append(QString(" %1").arg(chan->limit()));

  __ModControl->sendnotice(sender, QString("Informations about channel %1").arg(chan->name()));
  __ModControl->sendnotice(sender, QString("Creation timestamp : %1").arg(chan->creationTS().toString("dd/MM/yy hh:mm:ss")));
  if (!strmode.isEmpty())
    __ModControl->sendnotice(sender, QString("Mode flags         : +%1").arg(strmode));
  if (!chan->topic().isEmpty())
    __ModControl->sendnotice(sender, QString("Channel topic      : %1").arg(chan->topic()));

  QHash<unsigned int, MChanMember *> members = chan->memberList();
  QHash<unsigned int, MChanMember *>::iterator it2;
  int membercount = 0, opcount = 0;
  for(it2 = members.begin(); it2 != members.end(); ++it2) {
    MChanMember *member = it2.value();
    if (!member)
      continue;

    QString strprefix;
    if (FLAG_SET(member->flags(), MChanMember::Flag_ChanOp)) {
      strprefix = "@";
      opcount++;
    } else if (FLAG_SET(member->flags(), MChanMember::Flag_Voice)) {
      strprefix = "+";
    }

    __ModControl->sendnotice(sender, QString("%1. %2%3 (%4@%5)")
                                       .arg(++membercount)
                                       .arg(strprefix)
                                       .arg(member->user()->nick())
                                       .arg(member->user()->ident())
                                       .arg(member->user()->hostname()));
  }

  __ModControl->sendnotice(sender, QString("Total %1 members (%2 ops) in the channel").arg(membercount).arg(opcount));
  __ModControl->sendnotice(sender, QString("The channel has %1 ban(s) set").arg(chan->banCount()));
  __ModControl->sendnotice(sender, QString("End of channel info."));
}

void ModControl::c_stats(MUser *sender, const QStringList &strlArgs) {
  if (strlArgs.size() == 0)
    return;

  quint64 memUsed;
  memUsed = channelList.capacity() * sizeof(MChannel);
  memUsed+= userList.capacity() * sizeof(MUser);
  memUsed+= nickList.capacity() * (sizeof(MUser *) * 2);
  memUsed+= serverList.capacity() * sizeof(MServer);

  __ModControl->sendnotice(sender, QString("Channel hash table size : %1 (%2 buckets allocated)")
                                     .arg(channelList.size())
                                     .arg(channelList.capacity()));
  __ModControl->sendnotice(sender, QString("User hash table size    : %1 (%2 buckets allocated)")
                                     .arg(userList.size())
                                     .arg(userList.capacity()));
  __ModControl->sendnotice(sender, QString("Nick hash table size    : %1 (%2 buckets allocated)")
                                     .arg(nickList.size())
                                     .arg(nickList.capacity()));
  __ModControl->sendnotice(sender, QString("Server hash list size   : %1")
                                     .arg(serverList.size()));
  __ModControl->sendnotice(sender, QString("Memory usage            : %1 bytes")
		                             .arg(memUsed));
}

void ModControl::c_status(MUser *sender, const QStringList &strlArgs) {
  if (strlArgs.size() == 0)
    return;

  QApplication *app = globalMain->getApplication();
  __ModControl->sendnotice(sender, QString("%1, Version %2 (libmscore build date %3)").arg(app->applicationName()).arg(app->applicationVersion()).arg(globalMain->getBuilt()));
  __ModControl->sendnotice(sender, QString("MSloader path: %1").arg(app->applicationFilePath()));
  __ModControl->sendnotice(sender, QString("Boot time: %1 (PID %2)").arg(globalMain->getBootTime().toString("dd.MM.yyyy hh:mm:ss")).arg(app->applicationPid()));
  __ModControl->sendnotice(sender, QString("Socket: %1 kB in, %2 kB out").arg(globalMain->socket->getBytesRx() * 1024.0, 0, 'g', 1).arg(globalMain->socket->getBytesTx() * 1024.0, 0, 'g', 1));
}

void ModControl::c_whois(MUser *sender, const QStringList &strlArgs) {
  if (strlArgs.size() < 2) {
    __ModControl->sendnotice(sender, QString("Usage: whois nick"));
    return;
  }

  MUser *user;
  if (!(user = MUser::findByNick(strlArgs.at(1)))) {
    __ModControl->sendnotice(sender, QString("Unknown user %1.").arg(strlArgs.at(1)));
    return;
  }

  QString modestr = MUser::buildModeflagstr(user->modeflags());
  if (modestr.isEmpty())
    modestr = QString("(none)");

  __ModControl->sendnotice(sender, QString("Nick: %1 (%2@%3)").arg(user->nick()).arg(user->ident()).arg(user->hostname()));
  __ModControl->sendnotice(sender, QString("Connected at %1").arg(user->connectedAt().toString("dd/MM/yy hh:mm:ss")));
  __ModControl->sendnotice(sender, QString("Modes: %1").arg(modestr));
  __ModControl->sendnotice(sender, QString("Numeric: %1 (Base64: %2)").arg(user->numeric()).arg(MP10Parser::inttobase64(user->numeric(), 5)));
  if (!user->account().isEmpty())
    __ModControl->sendnotice(sender, QString("Authed as %1").arg(user->account()));
  if (user->server())
    __ModControl->sendnotice(sender, QString("Server: %1 (%2)").arg(user->server()->name()).arg(user->server()->description()));

  QList<MChannel *> channelList = user->channelList();
  QStringList strlChans;
  foreach(MChannel *chan, channelList)
    strlChans << chan->name();

  if (strlChans.size() > 0)
    __ModControl->sendnotice(sender, QString("Channels: %1").arg(strlChans.join(", ")));

  __ModControl->sendnotice(sender, QString("End of whois."));
}
