/**
 * ModServ IRC Services
 * Version 2009
 * Copyright (C) 2007, 2008 Christian Jurk <commx@commx.ws>
 * All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 3 of the License,
 * or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <http://www.gnu.org/licenses/>.
 */

#include <time.h>
#include <MS_P10_Parser.h>
#include <MS_Main.h>
#include <MS_Channels.h>
#include <MS_Protolib.h>
#include <MS_Servers.h>
#include <MS_Users.h>

MP10Parser::MP10Parser() {
  connect(globalMain->socket, SIGNAL(readyToParse(const QString &)), this, SLOT(splitLine(const QString &)));
  connect(globalMain->socket, SIGNAL(readyToAuth()), this, SLOT(initAuth()));
  
  connect(this, SIGNAL(onServerMessage(const QStringList &)), this, SLOT(handleServerMessage(const QStringList &)));
  connect(this, SIGNAL(onServerQuitMessage(const QStringList &)), this, SLOT(handleServerQuitMessage(const QStringList &)));
  connect(this, SIGNAL(onEndOfBurstMessage(const QStringList &)), this, SLOT(handleEndOfBurstMessage(const QStringList &)));
  connect(this, SIGNAL(onEndOfBurstAckMessage(const QStringList &)), this, SLOT(handleEndOfBurstAckMessage(const QStringList &)));
  connect(this, SIGNAL(onPingMessage(const QStringList &)), this, SLOT(handlePingMessage(const QStringList &)));
  connect(this, SIGNAL(onNickMessage(const QStringList &)), this, SLOT(handleNickMessage(const QStringList &)));
  connect(this, SIGNAL(onBurstMessage(const QStringList &)), this, SLOT(handleBurstMessage(const QStringList &)));
  connect(this, SIGNAL(onClearmodeMessage(const QStringList &)), this, SLOT(handleClearmodeMessage(const QStringList &)));
  connect(this, SIGNAL(onCreateMessage(const QStringList &)), this, SLOT(handleCreateMessage(const QStringList &)));
  connect(this, SIGNAL(onGlineMessage(const QStringList &)), this, SLOT(handleGlineMessage(const QStringList &)));
  connect(this, SIGNAL(onJoinMessage(const QStringList &)), this, SLOT(handleJoinMessage(const QStringList &)));
  connect(this, SIGNAL(onKickMessage(const QStringList &)), this, SLOT(handleKickMessage(const QStringList &)));
  connect(this, SIGNAL(onKillMessage(const QStringList &)), this, SLOT(handleKillMessage(const QStringList &)));
  connect(this, SIGNAL(onModeMessage(const QStringList &)), this, SLOT(handleModeMessage(const QStringList &)));
  connect(this, SIGNAL(onPartMessage(const QStringList &)), this, SLOT(handlePartMessage(const QStringList &)));
  connect(this, SIGNAL(onPrivmsgMessage(const QStringList &)), this, SLOT(handlePrivmsgMessage(const QStringList &)));
  connect(this, SIGNAL(onQuitMessage(const QStringList &)), this, SLOT(handleQuitMessage(const QStringList &)));
  connect(this, SIGNAL(onTopicMessage(const QStringList &)), this, SLOT(handleTopicMessage(const QStringList &)));
  
  MUser::init();
  MChannel::init();
}

void MP10Parser::splitLine(const QString &strBuffer) {
  strlParams = strBuffer.trimmed().split(" ");
  
  for(int i = 0; i < strlParams.size(); i++) {
    if (strlParams.at(i).left(1) == QString(":")) {
      QString strLongArg = strlParams.at(i).mid(1); /* just skip the first character */
      
      /* now add all following params into the long arg */
      for(int j = i + 1; j < strlParams.size(); j++)
        strLongArg += QString(" %1").arg(strlParams.at(j));

      int parc = strlParams.size();
      for(int k = i + 1; k < parc; k++)
        strlParams.removeLast();

      strlParams.replace(i, strLongArg);
      break;
    }
  }
  
  parseMessage();
}

void MP10Parser::initAuth(void) {
  MTcpSocket *sock = globalMain->socket;
  MConfigReader *cfg = globalMain->config;
  
  cleanup(); /* remove all existing servers, channels, users, etc. */
  m_localNum = inttobase64(cfg->get("vserver.numeric", "1").toUInt(), 2);
  
  sock->send(QString("PASS :%1").arg(cfg->get("server.pass", "password")));
  sock->send(QString("SERVER %1 1 %2 %3 J10 %4]]] +s :%5").arg(cfg->get("vserver.name", "modserv.example.org"))
               .arg(static_cast<quint32>(time(NULL)))
               .arg(static_cast<quint32>(time(NULL)))
               .arg(m_localNum)
               .arg(cfg->get("vserver.description", "ModServ IRC Services")));
  sock->send(QString("%1 EB").arg(m_localNum));
  
  MServer *srv = new MServer();
  srv->setServerName(cfg->get("vserver.name", "modserv.example.org"));
  srv->setServerTSBooted(globalMain->getBootTime());
  srv->setServerTSLinked(QDateTime::currentDateTime());
  srv->setServerNumeric(base64toint(m_localNum.toStdString().c_str()));
  srv->setServerFlags(MServer::FL_SERVICES);
  srv->setServerState(MServer::STATE_LINKED);
  srv->setServerDescription(cfg->get("vserver.description", "ModServ IRC Services"));
  virtualServer = srv;
  serverList.insert(srv->numeric(), srv);
}

void MP10Parser::cleanup(void) {
  QHash<unsigned int, MServer *>::iterator si;
  for(si = serverList.begin(); si != serverList.end(); ++si) {
    MServer *srv = si.value();
    delete srv;
  }
  serverList.clear();
  
  QHash<QString, MChannel *>::iterator ci;
  for(ci = channelList.begin(); ci != channelList.end(); ++ci) {
    MChannel *chan = ci.value();

    QHash<unsigned int, MChanMember *> hash = chan->memberList();
    QHash<unsigned int, MChanMember *>::iterator mi;
    for(mi = hash.begin(); mi != hash.end(); ++mi) {
      MChanMember *member = mi.value();
      delete member;
    }
  }
  
  QHash<unsigned int, MUser *>::iterator ui;
  for(ui = userList.begin(); ui != userList.end(); ++ui) {
    MUser *user = ui.value();
    delete user;
  }
  userList.clear();
}

void MP10Parser::parseMessage(void) {
  int pos = 1;
  
  if (strlParams.size() == 0)
    return; /* invalid line received */
  else if (strlParams.at(0) == QString("PASS") || strlParams.at(0) == QString("SERVER"))
    pos = 0; /* a few commands like PASS and SERVER from uplink */

  QString cmd = strlParams.at(pos);
  if (cmd == QString("B"))
    emit onBurstMessage(strlParams);
  else if (cmd == QString("C"))
    emit onCreateMessage(strlParams);
  else if (cmd == QString("D"))
    emit onKillMessage(strlParams);
  else if (cmd == QString("CM"))
    emit onClearmodeMessage(strlParams);
  else if (cmd == QString("EA"))
    emit onEndOfBurstAckMessage(strlParams);
  else if (cmd == QString("EB"))
    emit onEndOfBurstMessage(strlParams);
  else if (cmd == QString("G"))
    emit onPingMessage(strlParams);
  else if (cmd == QString("GL"))
    emit onGlineMessage(strlParams);
  else if (cmd == QString("J"))
    emit onJoinMessage(strlParams);
  else if (cmd == QString("K"))
    emit onKickMessage(strlParams);
  else if (cmd == QString("L"))
    emit onPartMessage(strlParams);
  else if (cmd == QString("M"))
    emit onModeMessage(strlParams);
  else if (cmd == QString("N"))
    emit onNickMessage(strlParams);
  else if (cmd == QString("OM"))
    emit onModeMessage(strlParams);
  else if (cmd == QString("P"))
    emit onPrivmsgMessage(strlParams);
  else if (cmd == QString("Q"))
    emit onQuitMessage(strlParams);
  else if (cmd == QString("SERVER") || cmd == QString("S"))
    emit onServerMessage(strlParams);
  else if (cmd == QString("SQ"))
    emit onServerQuitMessage(strlParams);
  else if (cmd == QString("T"))
    emit onTopicMessage(strlParams);
}

unsigned int MP10Parser::base64toint(const QString &strNum) {
  const char *s = strNum.toStdString().c_str();
  unsigned int i = convert2n[(unsigned char) *s++];
  while (*s) {
    i <<= NUMNICKLOG;
    i += convert2n[(unsigned char) *s++];
  }
  return i;
}

const QString MP10Parser::inttobase64(unsigned int v, unsigned int count) {
  char buf[32];
  
  if (count > 32)
    return QString("");
    
  buf[count] = '\0';
  while (count > 0) {
    buf[--count] = convert2y[(v & NUMNICKMASK)];
    v >>= NUMNICKLOG;
  }
  
  return QString(buf);
}

/* Server message handler
 * XX S servername hops boot_ts link_ts protocol numeric/maxusers flags :description
 */
void MP10Parser::handleServerMessage(const QStringList &strlArgs) {
  int pos = 1;
  MServer *srv;
  
  if (strlArgs.at(0) != QString("SERVER"))
    pos++;

   // SERVER name hops booted linked protocol numeric flags :desc
  if (!(srv = findServerByName(strlArgs.at(pos)))) {
    unsigned short flags = 0;
    srv = new MServer;
    
    srv->setServerName(strlArgs.at(pos));
    srv->setServerTSBooted(QDateTime::fromTime_t(strlArgs.at(pos + 2).toUInt()));
    srv->setServerTSLinked(QDateTime::fromTime_t(strlArgs.at(pos + 3).toUInt()));
    srv->setServerNumeric(base64toint(strlArgs.at(pos + 5).left(2).toStdString().c_str()));
    
    QString strFlags = strlArgs.at(pos + 6);
    if (strFlags.contains("h"))
      flags |= MServer::FL_HIDDEN;
    if (strFlags.contains("s"))
      flags |= MServer::FL_SERVICES;

    srv->setServerFlags(flags);
    srv->setServerDescription(strlArgs.at(pos + 7));
    
    serverList.insert(srv->numeric(), srv);
    
    if (pos == 1) /* it's my uplink */
      uplinkServer = srv;
    else
      srv->setServerState(MServer::STATE_LINKED);
    
    globalMain->log->print(L_NOTICE, QString("Adding server: %1%2").arg(srv->name()).arg(uplinkServer == srv ? QString(" (uplink)") : QString("")));
  }
}

/* Server quit message handler
 * XX SQ servername ts [:reason]
 */
void MP10Parser::handleServerQuitMessage(const QStringList &strlArgs) {
  MServer *srv;
  
  if ((srv = findServerByName(strlArgs.at(2)))) {
    if (srv->linkTS() != QDateTime::fromTime_t(strlArgs.at(3).toUInt()))
      return; /* link timestamps do not match, ignore */

    globalMain->log->print(L_NOTICE, QString("Deleting server: %1").arg(srv->name()));
    
    foreach(MUser *user, srv->userList) {
      QList<MChannel *> chList = user->channelList();
      
      foreach(MChannel *chan, chList) {
        chan->removeMember(user);
      }
      
      delete user;
    }
    
    serverList.remove(srv->numeric());
    delete srv;
  }
}

/* End of burst message handler
 * XX EB
 */
void MP10Parser::handleEndOfBurstMessage(const QStringList &strlArgs) {
  MServer *srv;

  if (!(srv = findServer(strlArgs.at(0)))) {
    globalMain->log->print(L_WARN, QString("Received END_OF_BURST from non-existing server (numeric %1)").arg(strlArgs.at(0)));
    return;
  }
  
  if (srv->state() >= MServer::STATE_END_OF_BURST) {
    globalMain->log->print(L_WARN, QString("Received END_OF_BURST from %1 (already bursted)").arg(srv->name()));
    return;
  }
  
  MTcpSocket *sock = globalMain->socket;
  sock->send(QString("%1 EA").arg(m_localNum));
  
  srv->setServerState(MServer::STATE_END_OF_BURST);
}

/* End of burst Ack message handler
 * XX EA
 */
void MP10Parser::handleEndOfBurstAckMessage(const QStringList &strlArgs) {
  MServer *srv;

  if (!(srv = findServer(strlArgs.at(0)))) {
    globalMain->log->print(L_WARN, QString("Received END_OF_BURST_ACK from non-existing server (numeric %1)").arg(strlArgs.at(0)));
    return;
  }

  if (srv->state() >= MServer::STATE_LINKED) {
    globalMain->log->print(L_WARN, QString("Received END_OF_BURST_ACK from %1 (link already completed)").arg(srv->name()));
    return;
  }

  if (uplinkServer && uplinkServer == srv)
    globalMain->log->print(L_NOTICE, QString("net.burst completed with uplink %1").arg(uplinkServer->name()));

  srv->setServerState(MServer::STATE_LINKED);
}

/* Ping message handler
 * XX G random servername random
 */
void MP10Parser::handlePingMessage(const QStringList &strlArgs) {
  if (virtualServer && strlArgs.at(3) == virtualServer->name()) {
    MTcpSocket *sock = globalMain->socket;
    sock->send(QString("%1 Z :%2").arg(m_localNum).arg(virtualServer->name()));
  }
}

/* Nick message handler
 * XX N nick hops ident host [+modes ...] numeric base64ip :realname
 */
void MP10Parser::handleNickMessage(const QStringList &strlArgs) {
  MServer *srv;
  MUser *user;
  
  globalMain->log->print(L_WARN, QString("Debug: handleNickMessage"));
  
  if (strlArgs.size() == 4) {
    globalMain->log->print(L_WARN, QString("Debug: nick change"));
    if ((user = MUser::find(base64toint(strlArgs.at(0))))) {
      nickList.remove(user->nick().toLower());
      nickList.insert(strlArgs.at(2).toLower(), user);
      user->setNick(strlArgs.at(2));
      globalMain->log->print(L_WARN, QString("Debug: nickchange done"));
    }
  } else if (!(user = MUser::findByNick(strlArgs.at(2)))) { /* new nick */
    globalMain->log->print(L_WARN, QString("Debug: new nick"));
    if (!(srv = findServer(strlArgs.at(0)))) {
      globalMain->log->print(L_WARN, QString("Received NICK message from non-existing server (numeric %1)").arg(strlArgs.at(0)));
      return;
    }
  
    globalMain->log->print(L_WARN, QString("Debug: server found"));
    quint32 modeflags = 0;
    user = new MUser;
    user->setNick(strlArgs.at(2));
    user->setConnectedAt(QDateTime::fromTime_t(strlArgs.at(4).toUInt()));
    user->setIdent(strlArgs.at(5));
    user->setHostname(strlArgs.at(6));
    
    globalMain->log->print(L_WARN, QString("Debug: collecting modes"));
    if (strlArgs.at(7).at(0) == '+') {
      int pc = 8;
      char c;
      for(int i=1; i < strlArgs.at(7).length(); i++) {
        c = strlArgs.at(7).at(i).toAscii();
        if (umodeList.contains(c))
          FLAG_ADD(modeflags, umodeList.value(c));

        /* parse mode args */
        switch(c) {
          case 'r':
            user->setAccount(strlArgs.at(pc++));
            break;
          case 'h':
            /* quakenet extension */
            break;
          default:
            break;
        }
      }

      user->setModeflags(modeflags);
    }
    
    globalMain->log->print(L_WARN, QString("Debug: settings server"));
    user->setServer(srv);
    user->setNumeric(base64toint(strlArgs.at(strlArgs.size() - 2)));
    userList.insert(user->numeric(), user);
    nickList.insert(user->nick().toLower(), user);
    globalMain->log->print(L_WARN, QString("Debug: nick intro done"));
  } else if (strlArgs.size() >= 9) { /* check for nick collision */
    globalMain->log->print(L_WARN, QString("Debug: check for collision"));
    QDateTime stamp = QDateTime::fromTime_t(strlArgs.at(4).toUInt());
    bool maskMatches = (user->ident() == strlArgs.at(5) && user->hostname() == strlArgs.at(6)) ? true : false;
    QString reason = QString("Nick collision");
    
    if (stamp == user->nicktime()) { /* kill both */
      killuser(user, reason);
      killnumeric(strlArgs.at(strlArgs.size() - 2), reason);
    } else if (stamp > user->nicktime() && !maskMatches) { /* kill the younger */
      killnumeric(strlArgs.at(strlArgs.size() - 2), reason);
    } else if (stamp < user->nicktime() && maskMatches) { /* kill the older */
      killuser(user, reason);
    }
  }
}

/* Burst handler
 * XX B #channel creationts [+modes ...] userlist [:%bans]
 */
void MP10Parser::handleBurstMessage(const QStringList &strlArgs) {
  int upa = 4; /* userlist start parameter */
  bool ignoreFlags = false;
  MChannel *chan;
  
  /* check for an existing channel */
  if (channelList.contains(strlArgs.at(2))) {
    chan = channelList.value(strlArgs.at(2));
    if (chan->creationTS() > QDateTime::fromTime_t(strlArgs.at(3).toUInt())) {
      /* clear topic/modes/ops/voices/bans */
      chan->setModeflags(0);
      chan->setTopic("");
      
      QHash<unsigned int, MChanMember *> hash = chan->memberList();
      QHash<unsigned int, MChanMember *>::iterator mi;
      for(mi = hash.begin(); mi != hash.end(); ++mi) {
        MChanMember *member = mi.value();
        member->setFlags(0);
      }
      
      chan->clearBanlist();
    } else if (chan->creationTS() < QDateTime::fromTime_t(strlArgs.at(3).toUInt())) {
      ignoreFlags = true;
    }
  }
  
  chan = new MChannel(strlArgs.at(2));
  chan->setCreationTS(QDateTime::fromTime_t(strlArgs.at(3).toUInt()));
  
  if (strlArgs.size() > 4) {
    /* parse channel modes */
    if (strlArgs.at(4).at(0) == QChar('+')) {
      quint32 flags = 0;
    
      if (!ignoreFlags) {
        for(int i=1; i < strlArgs.at(4).size(); i++) {
          char c = strlArgs.at(4).at(i).toAscii();
      
          if (cmodeList.contains(c))
            FLAG_ADD(flags, cmodeList.value(c));

          if (c == 'k')
            chan->setKey(strlArgs.at(++upa));
          else if (c == 'l')
            chan->setLimit(strlArgs.at(++upa).toUInt());
        }
      }
      
      chan->setModeflags(flags);
    
      ++upa;
    }
    
    /* parse channel members and their flags */
    QStringList strlMembers;
    quint32 flags = 0;
    unsigned int numeric;
    MUser *user;

    if (strlArgs.at(upa).contains(","))
      strlMembers = strlArgs.at(upa).split(",");
    else
      strlMembers << strlArgs.at(upa);

    for(int i=0; i < strlMembers.size(); i++) {
      QString num = strlMembers.at(i);
      
      if (num.length() > 5 && num.at(5) == QChar(':')) {
        QString flagstr = num.mid(5);
        flags = 0;
        
        for(int j=0; j < flagstr.length(); j++) {
          switch(flagstr.at(j).toAscii()) {
            case 'o':
              FLAG_ADD(flags, MChanMember::Flag_ChanOp);
              break;
            case 'v':
              FLAG_ADD(flags, MChanMember::Flag_Voice);
              break;
            default:
              break;
          }
        }
        
        num = num.left(5);
      }
      
      numeric = base64toint(num);
      if (!(user = MUser::find(numeric)))
        continue; /* user not found?! */
        
      chan->addMember(user, (ignoreFlags) ? 0 : flags);
    }
    
    /* look if we have channel bans */
    if (upa + 1 < strlArgs.size() && strlArgs.at(upa + 1).at(0) == QChar('%')) {
      QStringList strlBans;
      
      if (strlArgs.at(upa + 1).contains(" "))
        strlBans = strlArgs.at(upa + 1).split(" ");
      else
        strlBans << strlArgs.at(upa + 1);

      for(int i=0; i < strlBans.size(); i++) {
        QString strBanmask = strlBans.at(i);
        if (strBanmask.at(0) == QChar('%'))
          strBanmask = strBanmask.mid(1);

        if (!ignoreFlags)
          chan->addBan(strBanmask);
      }
    }
  }
  
  channelList.insert(chan->name().toLower(), chan);
}

/* Clearmode handler
 * XX CM #channel :flags
 */
void MP10Parser::handleClearmodeMessage(const QStringList &strlArgs) {
  MChannel *chan;
  quint32 flags = 0;
  
  if (!(chan = MChannel::find(strlArgs.at(2))))
    return; /* unable to find the channel */

  flags = chan->modeflags();
  
  for(int i=0; i < strlArgs.at(3).length(); i++) {
    char c = strlArgs.at(3).at(i).toAscii();
    
    if (cmodeList.contains(c)) {
      if (flags & cmodeList.value(c)) {
        FLAG_DEL(flags, cmodeList.value(c));
      }
    } else if (c == 'o' || c == 'v') {
      QHash<unsigned int, MChanMember *> hash = chan->memberList();
      QHash<unsigned int, MChanMember *>::iterator mi;
      for(mi = hash.begin(); mi != hash.end(); ++mi) {
        MChanMember *member = mi.value();
        member->setFlags(0);
      }
    } else if (c == 'b') {
      chan->clearBanlist();
    }
  }
  
  chan->setModeflags(flags);
}

/* Create handler
 * XXXXX C #channel creationTS
 */
void MP10Parser::handleCreateMessage(const QStringList &strlArgs) {
  MTcpSocket *sock = globalMain->socket;
  MChannel *chan;
  MUser *user;
  QStringList strlChannels;
  QDateTime TS = QDateTime::fromTime_t(strlArgs.at(3).toUInt());
  QDateTime TS_LagTime = QDateTime::currentDateTime();
  quint32 flags = 0;
  
  TS_LagTime.addSecs(3600);
  
  if (!(user = MUser::find(base64toint(strlArgs.at(0))))) {
    globalMain->log->print(L_WARN, QString("CREATE from non-existant user %1 (%2)").arg(strlArgs.at(0)).arg(strlArgs.at(2)));
    return;
  }

  if (strlArgs.at(2).contains(","))
    strlChannels = strlArgs.at(2).split(",");
  else
    strlChannels << strlArgs.at(2);

  foreach(QString str, strlChannels) {
    if ((chan = MChannel::find(str))) { /* channel exists already */
      if (chan->creationTS() > TS || TS > TS_LagTime) {
        sock->send(QString("%1 M %2 -o %3").arg(m_localNum)
                     .arg(chan->name())
                     .arg(strlArgs.at(0)));
        flags = 0;
      } else if (chan->creationTS() < TS) {
        /* there is no support for MAGIC_REMOTE_JOIN_TS, as it should not occur on P10 servers.
           it was implemented for P9 compability */
        chan->setCreationTS(TS);
        FLAG_ADD(flags, MChanMember::Flag_ChanOp);
      } else {
        FLAG_ADD(flags, MChanMember::Flag_ChanOp);
      }
    } else { /* channel is created */
      chan = new MChannel(str);
      chan->setCreationTS(TS);
      channelList.insert(str, chan);
      FLAG_ADD(flags, MChanMember::Flag_ChanOp);
    }
    
    chan->addMember(user, flags);
  }
}

/* GLine handler
 * XX GL target [+|-]mask duration creationTS :reason
 */
void MP10Parser::handleGlineMessage(const QStringList &strlArgs) {
  if (strlArgs.size() == 0)
    return; /* just a warning hide */
}

/* Join handler
 * XXXXX J channel(s) [TS]
 */
void MP10Parser::handleJoinMessage(const QStringList &strlArgs) {
  MUser *user;
  MChannel *chan;
  QStringList strlChannels;
  
  if (!(user = MUser::find(base64toint(strlArgs.at(0))))) {
    globalMain->log->print(L_WARN, QString("JOIN from non-existant user %1 (%2)").arg(strlArgs.at(0)).arg(strlArgs.at(2)));
    return;
  }
  
  if (strlArgs.at(2) == QString("0")) {
    deleteUserFromAllChannels(user);
  } else {
    if (strlArgs.at(2).contains(","))
      strlChannels = strlArgs.at(2).split(",");
    else
      strlChannels << strlArgs.at(2);
      
    foreach(QString strchan, strlChannels) {
      if (!(chan = MChannel::find(strchan))) {
        chan = new MChannel(strchan);
        chan->setCreationTS(QDateTime::fromTime_t(strlArgs.at(3).toUInt()));
      } else {
        chan->addMember(user, 0);
      }
    }
  }
}

/* Kick handler
 * XXXXX K #channel numeric [:reason]
 */
void MP10Parser::handleKickMessage(const QStringList &strlArgs) {
  MUser *user;
  
  if (!(user = MUser::find(base64toint(strlArgs.at(3))))) {
    globalMain->log->print(L_WARN, QString("KICK non-existant user %1 from %2").arg(strlArgs.at(3)).arg(strlArgs.at(2)));
    return;
  }
  
  MChannel *chan;
  if (!(chan = MChannel::find(strlArgs.at(2)))) {
    globalMain->log->print(L_WARN, QString("KICK from non-existant channel %2 (user %1)").arg(strlArgs.at(2)).arg(user->nick()));
    return;
  }
  
  chan->removeMember(user);
}

/* Kill handler
 * XXXXX D numeric [:reason]
 */
void MP10Parser::handleKillMessage(const QStringList &strlArgs) {
  MUser *user;

  if (!(user = MUser::find(base64toint(strlArgs.at(3))))) {
    globalMain->log->print(L_WARN, QString("KILL non-existant user %1").arg(strlArgs.at(3)));
    return;
  }
  
  deluser(user);
}

/* Mode handler
 * XXXXX M nick [+|-]modes
 * XXXXX M #channel [+|-]modes
 */
void MP10Parser::handleModeMessage(const QStringList &strlArgs) {
  if (strlArgs.at(2).at(0) == QChar('#')) {
    MChannel *chan;
    MUser *user;
    MChanMember *member;
    bool adding = false;
    quint32 flags, ufl;
    int par = 3;
    
    if (!(chan = MChannel::find(strlArgs.at(2)))) {
      globalMain->log->print(L_WARN, QString("MODE for non-existant channel %1").arg(strlArgs.at(2)));
      return;
    }
    
    flags = chan->modeflags();
    
    for(int i=0; i < strlArgs.at(3).length(); i++) {
      QChar c = strlArgs.at(3).at(i);
      
      if (c == '+' || c == '-') {
        adding = (c == '+');
        continue;
      }
      
      if (adding && cmodeList.contains(c.toAscii())) {
        if (!FLAG_SET(flags, cmodeList.value(c.toAscii()))) {
          FLAG_ADD(flags, cmodeList.value(c.toAscii()));
        }
        
        switch(c.toAscii()) {
          case 'b':
            chan->addBan(strlArgs.at(++par));
            break;
            
          case 'k':
            chan->setKey(strlArgs.at(++par));
            break;
            
          case 'l':
            chan->setLimit(strlArgs.at(++par).toUInt());
            break;
            
          case 'o':
            if (!(user = MUser::find(base64toint(strlArgs.at(++par))))) {
              globalMain->log->print(L_WARN, QString("MODE +o for non-existant user %1 on %2").arg(strlArgs.at(par)).arg(chan->name()));
              break;
            }
            
            if (!(member = chan->member(user))) {
              globalMain->log->print(L_WARN, QString("MODE +o for user %1 not on chan %2").arg(user->nick()).arg(chan->name()));
              break;
            }
            
            ufl = member->flags();
            if (!FLAG_SET(ufl, MChanMember::Flag_ChanOp))
              FLAG_ADD(ufl, MChanMember::Flag_ChanOp);
              
            member->setFlags(ufl);
            break;
            
          case 'v':
            if (!(user = MUser::find(base64toint(strlArgs.at(++par))))) {
              globalMain->log->print(L_WARN, QString("MODE +v for non-existant user %1 on %2").arg(strlArgs.at(par)).arg(chan->name()));
              break;
            }

            if (!(member = chan->member(user))) {
              globalMain->log->print(L_WARN, QString("MODE +v for user %1 not on chan %2").arg(user->nick()).arg(chan->name()));
              break;
            }

            ufl = member->flags();
            if (!FLAG_SET(ufl, MChanMember::Flag_Voice))
              FLAG_ADD(ufl, MChanMember::Flag_Voice);

            member->setFlags(ufl);
            break;
            
          default:
            break;
        }
      } else if (cmodeList.contains(c.toAscii())) {
        if (FLAG_SET(flags, cmodeList.value(c.toAscii()))) {
          FLAG_DEL(flags, cmodeList.value(c.toAscii()));
        }
        
        switch(c.toAscii()) {
          case 'b':
            chan->removeBan(strlArgs.at(++par));
            break;

          case 'k':
            chan->setKey(QString(""));
            break;

          case 'l':
            chan->setLimit(0);
            break;

          case 'o':
            if (!(user = MUser::find(base64toint(strlArgs.at(++par))))) {
              globalMain->log->print(L_WARN, QString("MODE -o for non-existant user %1 on %2").arg(strlArgs.at(par)).arg(chan->name()));
              break;
            }

            if (!(member = chan->member(user))) {
              globalMain->log->print(L_WARN, QString("MODE -o for user %1 not on chan %2").arg(user->nick()).arg(chan->name()));
              break;
            }

            ufl = member->flags();
            if (FLAG_SET(ufl, MChanMember::Flag_ChanOp))
              FLAG_DEL(ufl, MChanMember::Flag_ChanOp);

            member->setFlags(ufl);
            break;

          case 'v':
            if (!(user = MUser::find(base64toint(strlArgs.at(++par))))) {
              globalMain->log->print(L_WARN, QString("MODE -v for non-existant user %1 on %2").arg(strlArgs.at(par)).arg(chan->name()));
              break;
            }

            if (!(member = chan->member(user))) {
              globalMain->log->print(L_WARN, QString("MODE -v for user %1 not on chan %2").arg(user->nick()).arg(chan->name()));
              break;
            }

            ufl = member->flags();
            if (FLAG_SET(ufl, MChanMember::Flag_Voice))
              FLAG_DEL(ufl, MChanMember::Flag_Voice);

            member->setFlags(ufl);
            break;

          default:
            break;
        }
      }
    }
    
    chan->setModeflags(flags);
  } else {
    MUser *user;
    
    if (!(user = MUser::findByNick(strlArgs.at(2)))) {
      globalMain->log->print(L_WARN, QString("MODE for non-existant user %1").arg(strlArgs.at(2)));
      return;
    }
    
    quint32 flags = user->modeflags();
    bool adding = false;
    
    for(int i=0; i < strlArgs.at(3).length(); i++) {
      QChar c = strlArgs.at(3).at(i);
      
      if (c == '+' || c == '-') {
        adding = (c == '+');
        continue;
      }
      
      if (adding) {
        if (umodeList.contains(c.toAscii())) {
          if (!FLAG_SET(flags, umodeList.value(c.toAscii()))) {
            FLAG_ADD(flags, umodeList.value(c.toAscii()));
          }
        }
      } else {
        if (umodeList.contains(c.toAscii())) {
          if (FLAG_SET(flags, umodeList.value(c.toAscii()))) {
            FLAG_DEL(flags, umodeList.value(c.toAscii()));
          }
        }
      }
    }
    
    user->setModeflags(flags);
  }
}

/* Part handler
 * XXXXX L #channel[,#channel2,...] [:reason]
 */
void MP10Parser::handlePartMessage(const QStringList &strlArgs) {
  MUser *user;
  MChannel *chan;
  QStringList strlChannels;
  
  if (!(user = MUser::find(base64toint(strlArgs.at(0))))) {
    globalMain->log->print(L_WARN, QString("PART from non-existant user %1 on %2").arg(strlArgs.at(0)).arg(strlArgs.at(2)));
    return;
  }
  
  if (strlArgs.at(2).contains(","))
    strlChannels = strlArgs.at(2).split(",");
  else
    strlChannels << strlArgs.at(2);
    
  foreach(QString strchan, strlChannels) {
    if (!(chan = MChannel::find(strchan))) {
      globalMain->log->print(L_WARN, QString("PART from non-existant channel %1 (user %2)").arg(strchan).arg(user->nick()));
      continue;
    } else {
      chan->removeMember(user);
    }
  }
}

/* Privmsg handler
 * XXXXX P target :message
 */
void MP10Parser::handlePrivmsgMessage(const QStringList &strlArgs) {
  MUser *sender;
  
  if (!(sender = MUser::find(base64toint(strlArgs.at(0))))) {
    globalMain->log->print(L_WARN, QString("PRIVMSG from non-existant user %1").arg(strlArgs.at(0)));
    return;
  }
  
  if (strlArgs.at(2).at(0) == QChar('#')) {
    // channel messages
  } else if (strlArgs.at(2).contains("@")) {
    // message to nick@server
  } else {
    // message to numeric
  }
}

/* Quit handler
 * XXXXX Q [:message]
 */
void MP10Parser::handleQuitMessage(const QStringList &strlArgs) {
  MUser *user;
  
  if (!(user = MUser::find(base64toint(strlArgs.at(0))))) {
    globalMain->log->print(L_WARN, QString("QUIT from non-existant user %1").arg(strlArgs.at(0)));
    return;
  }
  
  deluser(user);
}

/* Topic handler
 * XXXXX T #channel [channel creationtime] [topic changed time] :topic
 */
void MP10Parser::handleTopicMessage(const QStringList &strlArgs) {
  MChannel *chan;
  
  if (!(chan = MChannel::find(strlArgs.at(2)))) {
    globalMain->log->print(L_WARN, QString("TOPIC from non-existant channel %1").arg(strlArgs.at(2)));
    return;
  }
  
  chan->setTopic(strlArgs.at(strlArgs.size() - 1));
}
