/**
 * ModServ IRC Services
 * Version 2009
 * Copyright (C) 2007, 2008 Christian Jurk <commx@commx.ws>
 * All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 3 of the License,
 * or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <http://www.gnu.org/licenses/>.
 */

#include <MS_Users.h>
#include <MS_Main.h>
#include <MS_P10_Parser.h>

QHash<unsigned int, MUser *> userList;
QHash<QString, MUser *> nickList;
QHash<char, quint32> umodeList;

MUser::MUser(const QString &strNick) {
  m_modeflags = 0;
  setNick(strNick);
  setServer();
}

MUser::~MUser() {
  m_channelList.clear();
  
  if (m_server)
    m_server->userList.removeOne(this);
}

void MUser::setNick(const QString &strNick) {
  m_nick = strNick;
  setNicktime(QDateTime::currentDateTime());
}

void MUser::setIdent(const QString &strIdent) {
  m_ident = strIdent;
}

void MUser::setHostname(const QString &strHost) {
  m_host = strHost;
}

void MUser::setAccount(const QString &strAccount) {
  m_account = strAccount;
}

void MUser::setConnectedAt(const QDateTime &timestamp) {
  m_connected = timestamp;
}

void MUser::setNicktime(const QDateTime &timestamp) {
  m_nickTime = timestamp;
}

void MUser::setModeflags(const quint32 &flags) {
  m_modeflags = flags;
}

void MUser::setNumeric(const unsigned int &num) {
  m_numeric = num;
}

void MUser::setServer(MServer *srv) {
  m_server = srv;
  
  if (srv)
    srv->userList.append(this);
}

const QString &MUser::nick(void) {
  return m_nick;
}

const QString &MUser::ident(void) {
  return m_ident;
}

const QString &MUser::account(void) {
  return m_account;
}

const QString &MUser::hostname(void) {
  return m_host;
}

const QDateTime &MUser::connectedAt(void) {
  return m_connected;
}

const QDateTime &MUser::nicktime(void) {
  return m_nickTime;
}

const quint32 &MUser::modeflags(void) {
  return m_modeflags;
}

const unsigned int &MUser::numeric(void) {
  return m_numeric;
}

QString MUser::numeric64(void) {
  return MP10Parser::inttobase64(m_numeric, 5);
}

MServer *MUser::server(void) {
  return m_server;
}

void MUser::addChannel(MChannel *chan) {
  m_channelList.append(chan);
}

QList<MChannel *> MUser::channelList(void) {
  return m_channelList;
}

void MUser::removeChannel(MChannel *chan) {
  if (m_channelList.contains(chan)) {
    m_channelList.removeOne(chan);
  }
}

quint32 MUser::parseModeflags(const QString &strFlags) {
  quint32 flags = 0;
  
  for(int i=0; i < strFlags.length(); i++) {
    char c = strFlags.at(i).toAscii();
    if (umodeList.contains(c)) {
      FLAG_ADD(flags, umodeList.value(c));
    }
  }
  
  return flags;
}

QString MUser::buildModeflagstr(const quint32 &flags) {
  QString strFlags;
  QHash<char, quint32>::iterator it;
  
  for(it = umodeList.begin(); it != umodeList.end(); ++it) {
    if (FLAG_SET(flags, it.value()))
      strFlags.append(it.key());
  }
  
  return strFlags;
}

bool MUser::isAuthed(void) {
  return FLAG_SET(modeflags(), MUser::Flag_Account);
}

bool MUser::isOper(void) {
  return FLAG_SET(modeflags(), MUser::Flag_Operator);
}

void MUser::init(void) {
  if (umodeList.size() > 0)
    umodeList.clear();

  /* usermode flag 'a' (lowercase); boxenircu extension
   * Marks an user as network administrator.
   * This flag is available on kotaro and boxenircu only.
   */
  umodeList.insert('a', MUser::Flag_Admin);
  
  /* usermode flag 'd' (lowercase);
   * Marks an user as deaf; No channel conversations are received.
   * Useful for IRC services on large networks.
   */
  umodeList.insert('d', MUser::Flag_Deaf);
  
  /* usermode flag 'g' (lowercase);
   * Additional debug messages are received; requires usermode flag 'o' to be set.
   */
  umodeList.insert('g', MUser::Flag_Debug);
  
  /* usermode flag 'h' (lowercase); Quakenet extension
   * User used sethost to set a custom vhost.
   * Only available on ircds with asuka sethost-patch.
   * This command provides an additional parameter in NICK messages, the vhost.
   */
  umodeList.insert('h', MUser::Flag_Sethost);
  
  /* usermode flag 'i' (lowercase);
   * Marks an user invisible from /WHO, /NAMES, etc when not on channel.
   */
  umodeList.insert('i', MUser::Flag_Invisible);
  
  /* usermode flag 'k' (lowercase);
   * Marks an user as channel service. Users having this modeflag set are protected
   * from being deopped/kicked/killed.
   */
  umodeList.insert('k', MUser::Flag_Chanserv);
  
  /* usermode flag 'o' (lowercase);
   * Marks an user as IRC operator.
   */
  umodeList.insert('o', MUser::Flag_Operator);
  
  /* usermode flag 'r' (lowercase); u2.10.11 extension (& u2.10.10 + lain1.2_patches)
   * Sets the nework wide ACCOUNT name for the user.
   * This command provides an additional parameter in NICK messages, the account name.
   */
  umodeList.insert('r', MUser::Flag_Account);
  
  /* usermode flag 'R' (uppercase); Quakenet extension
   * Only authorised users can send PRIVMSG's and NOTICE's to the user.
   * Useful for blocking adverts/spam.
   */
  umodeList.insert('R', MUser::Flag_Noanonymous);
  
  /* usermode flag 'w' (lowercase);
   * User will receive wallop messages sent by IRC operators.
   */
  umodeList.insert('w', MUser::Flag_Wallops);
  
  /* usermode flag 'x' (lowercase); u2.10.11 extension
   * Sets an hidden host on the user when ACCOUNT is set.
   */
  umodeList.insert('x', MUser::Flag_Hiddenhost);
}

MUser *MUser::findByNick(const QString &strNick) {
  if (nickList.contains(strNick.toLower()))
    return nickList.value(strNick.toLower());
    
  return 0;
}

MUser *MUser::find(unsigned int num) {
  if (userList.contains(num))
    return userList.value(num);
    
  return 0;
}
